#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Use pyparser to parse the Nginx configuration file and construct
an instruction object.
"""

import codecs
import glob
import os
import re
import time
import pyparsing

pyparsing.ParserElement.setDefaultWhitespaceChars(" \n\t\r")
INTERVAL = float(0)


class NginxParserException(Exception):
    """
    Customized Configuration Parsing Exception
    """


class SimpleDirective:
    """
    A simple directive for Nginx configuration
    """
    def __init__(self, key=None, args=None, value=None, raw=None):
        self.key = key
        self.args = args
        self.value = value
        self.raw = raw

    def __repr__(self):
        repr_msg = "SimpleDirective(key=%r, args=%r, value=%r)" % (
            self.key,
            self.args,
            self.value,
        )
        return repr_msg

    def __str__(self):
        str_msg = "SimpleDirective(%s : %s -> %s)" % (
            self.key,
            self.args,
            self.value,
        )
        return str_msg


class BlockDirective(SimpleDirective):
    """
    A block directive for Nginx configuration
    """
    def __init__(self, key=None, args=None, value=None, raw=None):
        #super(BlockDirective, self).__init__(key, args, value, raw)
        SimpleDirective.__init__(self, key, args, value, raw)

    def __repr__(self):
        repr_msg = "BlockDirective(key=%r, args=%r, directives=%r)" % (
            self.key,
            self.args,
            self.value,
        )
        return repr_msg

    def __str__(self):
        str_msg = "BlockDirective(%s : %s -> %s)" % (
            self.key,
            self.args,
            self.value,
        )
        return str_msg


class IncludeNode:
    """
    Stores the path of the original include command and the parsed command information.
    """
    def __init__(self, file_name=None, model_name=None):
        self._file = file_name
        self._model = model_name

    def get_directives(self):
        """
        Return the parsed mode information.
        :return:
        """
        return self._model

    def get_file(self):
        """
        Return the file path information
        :return:
        """
        return self._file


class IncludeDirective(SimpleDirective):
    """
    Include Directive : special processing is required.
    The configuration file contained in include needs to be loaded during configuration parsing.
    """
    def __init__(self, key=None, args=None, value=None, raw=None):
        #super(IncludeDirective, self).__init__(key, args, value, raw)
        SimpleDirective.__init__(self, key, args, value, raw)
        self.extend = []
        self.abspath = None

    def __repr__(self):
        repr_msg = "IncludeDirective(key=%r, args=%r, value=%r, files=%r)" % (
            self.key,
            self.args,
            self.value,
            self.get_files(),
        )
        return repr_msg

    def __str__(self):
        str_msg = "IncludeDirective(%s : %s -> %s)" % (
            self.key,
            self.args,
            self.value,
        )
        return str_msg

    def set_abs_path(self, base_path):
        """
        Setting the root directory of the Nginx configuration File
        :param base_path: root directory
        :return: NA
        """
        if self.value is None:
            self.abspath = None
        else:
            abs_path = self.value

            if not os.path.isabs(abs_path):
                abs_path = os.path.join(base_path, abs_path)
            abs_path = os.path.realpath(abs_path)
            self.abspath = abs_path

    def add_extend(self, node_mode):
        """
        Add the parsed node.
        :param node_mode:
        :return:
        """
        if node_mode:
            self.extend.append(node_mode)

    def get_files(self):
        """
        Get file object
        :return: file list
        """
        file_head_list = []
        for file_head in self.extend:
            file_head_list.append(file_head.get_file())
        return file_head_list

    def get_abs_path(self):
        """
        Get root directory
        :return: root directory
        """
        return self.abspath

    def parse_include(self, base_path):
        """
        Parse the configuration file included in include.
        :param base_path: root directory
        :return: NA
        """
        abs_path = self.get_abs_path()

        # get all files under abs_path
        str_paths = glob.glob(abs_path)
        for str_path in str_paths:
            file_handle = None
            try:
                ret = load_from_file(str_path)
                if ret[0] != 0:
                    print("Load file %s failed:%s." % (str_path, ret[0]))
                    continue
                model = build_directives(ret[1], base_path=base_path)
                time.sleep(INTERVAL)
                self.add_extend(IncludeNode(str_path, model))
            except Exception:
                # If loading error,just only print error
                print("Failed to process include configuration file %s." %
                      str_path)
            finally:
                if file_handle:
                    file_handle.close()


def config_pyparser():
    """
    Set the pyparser rule and generate the pyparser parser.
    :return: pyparser parser object
    """
    space = pyparsing.Optional(pyparsing.White())
    non_space = pyparsing.Regex(r"\S+")
    left_bracket = pyparsing.Literal("{").suppress()
    right_bracket = space.leaveWhitespace() + pyparsing.Literal("}").suppress()
    semicolon = pyparsing.Literal(";").suppress()
    key = pyparsing.Word(pyparsing.alphanums + "_/+-.")

    condition = pyparsing.Regex(r"\(.+\)")

    double_quoted = pyparsing.Regex(r"(\".*\")")
    single_quoted = pyparsing.Regex(r"(\'.*\')")
    sub_vars = pyparsing.Regex(r"(\$\{\w+\})")
    nonspecial = pyparsing.Regex(r"[^\{\};,]")
    value_string = pyparsing.Combine(
        pyparsing.ZeroOrMore(double_quoted | single_quoted | sub_vars
                             | nonspecial))

    location = pyparsing.CharsNotIn("{};," + " \t\n\r\v\f")
    # matching rule for location directive
    # location [ = | ~ | ~* | ^~ ] uri { ... }
    location_matching = (pyparsing.Literal("=")
                         | pyparsing.Literal("~*")
                         | pyparsing.Literal("~")
                         | pyparsing.Literal("^~"))

    # Set parsing rules for Nginx configuration files
    comment = space + pyparsing.Literal("#") + pyparsing.restOfLine
    assignment = (space + key +
                  pyparsing.Optional(space + value_string, default=None) +
                  semicolon)
    location_statement = (space + pyparsing.Optional(location_matching) +
                          pyparsing.Optional(space + location + space))
    if_statement = space + pyparsing.Literal("if") + space + condition + space

    # rules for map directive
    # map $args $foo { }
    dollar_var = pyparsing.Combine(
        pyparsing.Literal("$") + pyparsing.Regex(r"[^\{\};,\s]+"))
    map_statement = (space + pyparsing.Literal("map") + space + non_space +
                     space + dollar_var + space)
    map_pattern = (pyparsing.Regex(r'".*"') | pyparsing.Regex(r"'.*'")
                   | non_space)
    map_entry = space + map_pattern + space + value_string + space + semicolon
    map_block = pyparsing.Group(
        pyparsing.Group(map_statement).leaveWhitespace() + left_bracket +
        pyparsing.Group(
            pyparsing.ZeroOrMore(pyparsing.Group(comment | map_entry)) +
            space).leaveWhitespace() + right_bracket)

    # rules for lua-linux-module
    key_lua = pyparsing.Regex("(.+_by_lua_block)|(.+_by_lua)")
    block_lua = pyparsing.Forward()
    block_lua_begin = (pyparsing.Group(space + key_lua +
                                       space)).leaveWhitespace()
    block_lua_innards = pyparsing.ZeroOrMore(pyparsing.Regex(r"[^\{\}]+"))
    block_lua << pyparsing.Group(block_lua_begin + left_bracket +
                                 block_lua_innards + right_bracket)

    # rules for limit_except GET POST
    block1 = pyparsing.Forward()
    block1_begin = pyparsing.Group(
        space + key +
        pyparsing.OneOrMore(pyparsing.Group(space + location + space)))
    block1_innards = pyparsing.Group(
        pyparsing.ZeroOrMore(pyparsing.Group(comment | assignment)) +
        space).leaveWhitespace()
    block1 << pyparsing.Group(block1_begin + left_bracket + block1_innards +
                              right_bracket)

    block = pyparsing.Forward()

    # Rules for normal block directive
    block_begin = (pyparsing.Group(space + key + location_statement)
                   ^ pyparsing.Group(if_statement)).leaveWhitespace()

    block_innards = pyparsing.Group(
        pyparsing.ZeroOrMore(block_lua
                             | pyparsing.Group(comment | assignment)
                             | block
                             | block1
                             | map_block) + space).leaveWhitespace()

    block << pyparsing.Group(block_begin + left_bracket + block_innards +
                             right_bracket)

    script = (pyparsing.OneOrMore(
        pyparsing.Group(comment | assignment) ^ block ^ map_block) + space +
              pyparsing.stringEnd)
    return script


def load_from_file(filename):
    """
    Read the configuration from the configuration file and
    use pyparser to parse the syntax.
    :param filename: config file name
    :return: List of parsed configuration information.
    """

    file_handle = None
    try:
        file_handle = codecs.open(filename)
        file_content = file_handle.read()
    except FileNotFoundError:
        return -1, "Nginx main config file is not exist."
    except Exception:
        return -1, "Failed to read content from config file."
    finally:
        if file_handle:
            file_handle.close()

    # remove *_by_lua_block|*_by_lua directives
    pattern = re.compile(
        r"\s+((\S+_by_lua_block)|(\S+_by_lua).+?';)",
        re.MULTILINE | re.IGNORECASE | re.S,
    )
    search_content_list = pattern.findall(file_content)
    for search_content in search_content_list:
        file_content = file_content.replace(search_content[0], "")

    # remove charset | charset_map directives
    p_match = re.compile(r"charset.+?;")
    file_content = p_match.sub("", file_content)

    script = config_pyparser()
    script.parseWithTabs().leaveWhitespace()
    content_list = script.parseString(file_content).asList()
    # remove space item in list
    content_list = remove_space(content_list)
    # remove comment items
    content_list = remove_comment(content_list)
    return 0, content_list


def set_interval(value):
    """
    Set the interval (in seconds) of internal task polling each other.
    :param value: interval value
    :return: NA
    """
    global INTERVAL
    if value >= 0:
        INTERVAL = value
    else:
        INTERVAL = float(0)


def remove_space(contents):
    """
    Remove blank items from the list(\n\t\r)
    :param contents: content list
    :return: content list
    """
    ret = []
    for entry in contents:
        if isinstance(entry, list):
            sublist = remove_space(entry)
            ret.append(sublist)
        elif (isinstance(entry, str) and entry.isspace()) or entry == "":
            pass
        else:
            ret.append(entry)

    return ret


def remove_comment(contents):
    """
    Remove blank items from the list(\n\t\r)
    :param contents: content list
    :return: content list
    """
    if not contents and len(contents) == 0:
        return contents

    ret = []

    if isinstance(contents, list):
        if isinstance(contents[0], str):
            if contents[0] != "#":
                return contents
        elif isinstance(contents[0], list):
            for sub in contents:
                r = remove_comment(sub)
                if r is not None:
                    ret.append(r)

    return None if len(ret) == 0 else ret


def build_key(cfg):
    """
    Return the key in the configuration file
    :param cfg: config list
    :return: key
    """
    key_name = None
    other_args = []

    if cfg is None:
        return key_name, other_args

    key_name = cfg
    if not isinstance(cfg, list):
        return key_name, other_args

    if len(cfg) <= 1:
        return key_name, other_args

    key_name = cfg[0]
    for args in cfg[1:]:
        if isinstance(args, list):
            other_args.extend(args)
        else:
            other_args.append(args)

    return key_name, other_args


def build_directives(config_content, base_path=None):
    """
    Modules that parse configuration files in detail
    :param config_content:config list
    :param base_path:
    :return:
    """
    # If not a list
    if not isinstance(config_content, list):
        raise NginxParserException("Nginx configuration file error")

    # Parse configuration into detailed modules
    root = BlockDirective(args=[], value=[], raw=config_content)

    for sub_line in config_content:

        if not isinstance(sub_line, list):
            raise NginxParserException("Directive expected: %s error" %
                                       sub_line)

        if len(sub_line) < 2:
            key, args = build_key(sub_line[0])
            sub_dir = SimpleDirective(key=key,
                                      args=args,
                                      value=None,
                                      raw=sub_line)
            root.value.append(sub_dir)
            continue

        sub_key = sub_line[0]
        sub_value = sub_line[1]
        if isinstance(sub_value, list):
            sub_block = build_directives(sub_value, base_path=base_path)
            key, args = build_key(sub_key)
            sub_block.key = key
            sub_block.args = args
            sub_block.raw = sub_line
            root.value.append(sub_block)
            continue

        key, args = build_key(sub_key)
        if isinstance(key, str):
            # Remove comment message
            if key[0] == "#":
                continue

            if key.lower() == "include":
                sub_dir = IncludeDirective(key=key,
                                           args=args,
                                           value=sub_value,
                                           raw=sub_line)
                sub_dir.set_abs_path(base_path)
                sub_dir.parse_include(base_path)
            else:
                sub_dir = SimpleDirective(key=key,
                                          args=args,
                                          value=sub_value,
                                          raw=sub_line)
            root.value.append(sub_dir)

    return root


def find_all_in_directives(directives, key):
    """
    Queries all directives whose keys are specified.
    :param directives: all directives
    :param key: directive key value
    :return: directives
    """
    return_value = []
    if isinstance(directives, list):
        target_data = directives
    else:
        target_data = directives.value

    for sub_data in target_data:
        if sub_data.key == key or sub_data.key == [key]:
            return_value.extend([sub_data])

        if isinstance(sub_data, BlockDirective):
            block_value = find_all_in_directives(sub_data, key)
            return_value.extend(block_value)

        if sub_data.key == "include" or sub_data.key == ["include"]:
            for sub in sub_data.extend:
                include_value = find_all_in_directives(sub.get_directives(),
                                                       key)
                return_value.extend(include_value)

    return return_value


def find_in_directives(directives, path):
    """
    Finds the elements defined by the array of paths in
    the configuration model. Search model can
    be BlockDirective or list
    :param directives: all directives
    :param path: like [http, include]
    :return: directives
    """
    return_value = []

    # If the search path is empty
    if path is None or len(path) == 0:
        return [directives]

    # Start finding blocks
    target_data = (directives
                   if isinstance(directives, list) else directives.value)

    for sub_data in target_data:
        if sub_data.key == "include" or sub_data.key == ["include"]:
            for sub in sub_data.extend:
                include_value = find_in_directives(sub.get_directives(),
                                                   path[0:])
                return_value.extend(include_value)

        elif sub_data.key == path[0] or sub_data.key == [path[0]]:
            if isinstance(sub_data, BlockDirective):
                block_value = find_in_directives(sub_data, path[1:])
                return_value.extend(block_value)
            elif isinstance(sub_data, SimpleDirective):
                return_value.extend([sub_data])
            else:
                raise NginxParserException("Not support directive type")

    return return_value
